unit XMLObject;

interface

uses classes, TypInfo, MSXML_TLB, contnrs;

type
  TXMLObject = class(TCollectionItem)
  private
    FPropList: PPropList;
    FPropCount: Integer;
    FObjectList: TObjectList;
    procedure InitPropList;
    procedure AddToObjectList(AObject: TObject);
  protected
    procedure SaveToStream(Stream: TStream);
  public
    destructor Destroy; override;
    procedure SaveToFile(const FileName: string);
    function xml: string;
  end;


function CreateXMLObjFromFile(const FileName: string): TXMLObject;
function CreateXMLObjFromString(const Text: string): TXMLObject;

implementation

uses Sysutils;

const
  tagXMLOBJECT = 'XMLOBJECT';
  attClassName = 'ClassName';
  attCollectionItem = 'CollectionItem';

function CreateObject(Element: IXMLDOMElement; ObjPers: TXMLObject): TXMLObject;
var
  Node: IXMLDOMElement;
  Nodes: IXMLDOMNode;
  Collection: TCollection;
  XMLObject: TXMLObject;
  CollectionItem: string;
begin
  // Vrifier que l'lment a comme nom XMLOBJECT et un attribut ClassName
  // ...

  if ObjPers = nil then
    ObjPers := FindClass(Element.attributes[0].text).Create as TXMLObject;

  ObjPers.InitPropList;

  Node := Element.firstChild as IXMLDOMElement;
  while Node <> nil do
  begin
    if IsPublishedProp(ObjPers, Node.nodeName) then
    begin
      if (Node.nodeType = NODE_ELEMENT) then
      begin
        if (Node.firstChild <> nil) and (Node.firstChild.nodeName = tagXMLOBJECT) then
        begin
          if (Node.firstChild as IXMLDOMElement).getAttribute(attCollectionItem) = Null then
          begin
            XMLObject := CreateObject(Node.firstChild as IXMLDOMElement, nil);
            ObjPers.AddToObjectList(XMLObject);
            SetObjectProp(ObjPers, Node.nodeName, XMLObject);
          end
          else
          begin
            CollectionItem := (Node.firstChild as IXMLDOMElement).getAttribute(attCollectionItem);
            Collection := TCollection.Create(TXMLObject);
            ObjPers.AddToObjectList(Collection);
            SetObjectProp(ObjPers, Node.nodeName, Collection);

            Nodes := Node.firstChild.firstChild;
            while Nodes <> nil do
            begin
              if Nodes.nodeType = NODE_ELEMENT then
              begin
                XMLObject := TCollectionItemClass(FindClass(CollectionItem)).Create(Collection) as TXMLObject;
                CreateObject(Nodes as IXMLDOMElement, XMLObject);
              end;
              Nodes := Nodes.nextSibling;
            end;
          end;
        end
        else
          case PropType(ObjPers, Node.nodeName) of
            tkInteger:
              SetOrdProp(ObjPers, Node.nodeName, StrToInt(Node.text));

            tkFloat:
              SetFloatProp(ObjPers, Node.nodeName, StrToFloat(Node.text));

            tkLString:
              SetStrProp(ObjPers, Node.nodeName, Node.text);
          end;
      end;
    end;
    Node := Node.nextSibling as IXMLDOMElement;
  end;

  Result := ObjPers;
end;

function CreateXMLObjFromFile(const FileName: String): TXMLObject;
var
  Doc: IXMLDOMDocument;
begin
  Doc := CoDOMDocument.Create;
  Doc.load(FileName);
  if Doc.parseError.errorCode <> 0 then
    raise Exception.Create(Doc.parseError.reason);

  Result := CreateObject(Doc.documentElement, nil);
end;

function CreateXMLObjFromString(const Text: string): TXMLObject;
var
  Doc: IXMLDOMDocument;
begin
  Doc := CoDOMDocument.Create;
  Doc.loadXML(Text);
  if Doc.parseError.errorCode <> 0 then
    raise Exception.Create(Doc.parseError.reason);

  Result := CreateObject(Doc.documentElement, nil);
end;

procedure WriteString(Stream: TStream; const Str: string);
var
  Pc: PChar;
begin
  Pc := PChar(Str);
  Stream.Write(Pc^, Length(Str));
end;


{ TXMLObject }

procedure TXMLObject.InitPropList;
var
  RTTIPtr: PTypeInfo;
begin
  if FPropList <> nil then Exit;

  RTTIPtr := ClassInfo;
  FPropCount := GetTypeData(RTTIPtr).PropCount;
  GetMem(FPropList, FPropCount * SizeOf(Pointer));
  GetPropInfos(RTTIPtr, FPropList);
end;

destructor TXMLObject.Destroy;
begin
  if FPropList <> nil then
    FreeMem(FPropList, FPropCount * SizeOf(Pointer));
  FObjectList.Free;

  inherited;
end;

procedure TXMLObject.AddToObjectList(AObject: TObject);
begin
  if FObjectList = nil then
    FObjectList := TObjectList.Create;
  FObjectList.Add(AObject);
end;

procedure TXMLObject.SaveToStream(Stream: TStream);
var
  I, J: Integer;
  Obj: TObject;
begin
  InitPropList;

  WriteString(Stream, Format('<%s %s="%s">', [tagXMLOBJECT, attClassName, Self.ClassName]));

  for I := 0 to FPropCount-1 do
    case FPropList[I].PropType^.Kind of
     tkInteger:
        WriteString(Stream, Format('<%s>%s</%s>', [
          FPropList[I].Name,
          IntToStr(GetOrdProp(Self, FPropList[I].Name)),
          FPropList[I].Name])
        );

      tkFloat:
        WriteString(Stream, Format('<%s>%s</%s>', [
          FPropList[I].Name,
          FloatToStr(GetFloatProp(Self, FPropList[I].Name)),
          FPropList[I].Name])
        );

      tkLString:
        WriteString(Stream, Format('<%s>%s</%s>', [
          FPropList[I].Name,
          GetStrProp(Self, FPropList[I].Name),
          FPropList[I].Name])
        );

      tkClass:
      begin
        Obj := GetObjectProp(Self, FPropList[I].Name);
        if Obj is TXMLObject then
        begin
          WriteString(Stream, Format('<%s>', [FPropList[I].Name]));
          TXMLObject(Obj).SaveToStream(Stream);
          WriteString(Stream, Format('</%s>', [FPropList[I].Name]));
        end
        else if Obj is TCollection then
        begin
          WriteString(Stream, Format('<%s>', [FPropList[I].Name]));
          with TCollection(Obj) do
          begin
            WriteString(Stream, Format('<%s %s="%s">',
              [tagXMLOBJECT, attCollectionItem, ItemClass.ClassName]));

            for J := 0 to Count-1 do
              if Items[J] is TXMLObject then
                TXMLObject(Items[J]).SaveToStream(Stream);

            WriteString(Stream, Format('</%s>', [tagXMLOBJECT]));
          end;
          WriteString(Stream, Format('</%s>', [FPropList[I].Name]));
        end;
      end;
    end;

  WriteString(Stream, Format('</%s>', [tagXMLOBJECT]));
end;

procedure TXMLObject.SaveToFile(const FileName: string);
var
  FileStream: TFileStream;
begin
  FileStream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(FileStream);
  finally
    FileStream.Free;
  end;
end;

function TXMLObject.xml: string;
var
  StringStream: TStringStream;
begin
  StringStream := TStringStream.Create('');
  try
    SaveToStream(StringStream);
    Result := StringStream.DataString;
  finally
    StringStream.Free;
  end;
end;

end.
